import abc

from wrightyrion.agent import Agent
from gembox.debug_utils import Debugger


class PageHandler(abc.ABC):
    """
    PageHandler is a base class for all page handlers.

    PageHandler is responsible for handling one type of page. Multiple PageHandlers can form up a wrightyrion agent.
    """
    page_type = None

    def __init__(self, agent: Agent, debug_tool: Debugger):
        assert isinstance(agent, Agent), f"agent must be a `wrightyrion.agent.Agent`, but {type(agent)}"
        self.agent = agent
        self.debug_tool = debug_tool

    def check_url(self, test_url: str = None) -> bool:
        """
        Check whether the current url is the expected type.

        :return: (bool) True when the url is acceptable
        """
        if self.page_type is None:
            return True
        else:
            return self._check_url(test_url=test_url)

    @abc.abstractmethod
    def _check_url(self, test_url: str) -> bool:
        """
        This method check whether the current url is the expected type.

        The expected type should be specified in class variable `page_type`.

        :param test_url: (str) the url to be tested
        :return: (bool) True when the url is acceptable
        """
        return True


__all__ = ['PageHandler']



import abc
import playwright.async_api
from gembox.debug_utils import Debugger


class Handler(abc.ABC):
    def __init__(self, page: playwright.async_api.Page, debug_tool: Debugger):
        """
        Initialize the Handler.

        :param page: (playwright.async_api.Page) the page
        :param debug_tool: (Debugger) the debug tool
        """
        self._page = page
        self._debug_tool = debug_tool

    @property
    def page(self) -> playwright.async_api.Page:
        return self._page

    @property
    def debug_tool(self) -> Debugger:
        return self._debug_tool


__all__ = ['Handler']



class SoupNotLoadedException(Exception):
    """Raised when soup is not loaded"""
    pass


class FailedToLoadWebpageException(Exception):
    """Raised when failed to load webpage"""
    pass


__all__ = ['SoupNotLoadedException', 'FailedToLoadWebpageException']



import abc
import pathlib

from bs4 import BeautifulSoup
from gembox.debug_utils import Debugger
from gembox.io import ensure_pathlib_path

from .exception import FailedToLoadWebpageException


class PageParser(abc.ABC):
    """
    The base class of all page parsers, built upon BeautifulSoup.

    Page Parser is responsible for reading webpages from local files, and parsing the webpages to get the information.

    After `load_webpage`, you can always access the `soup` property to get the BeautifulSoup object.
    """

    def __init__(self, debug_tool: Debugger = None, encoding="utf-8"):
        self._encoding = encoding
        self._debug_tool = debug_tool if debug_tool is not None else Debugger()
        self._file_path = None
        self._soup = None

    def load_webpage(self, file_path: (str, pathlib.Path)):
        """
        Load webpage from local file.

        :param file_path: (str, pathlib.Path) the path to the local file
        :return: (None)
        """
        self.debug_tool.info(f"[{self.__class__.__name__}] Loading webpage from {file_path}...")
        try:
            self._file_path = ensure_pathlib_path(file_path)
            self._soup = self._read_webpage_from_file(file_path=file_path)
            self.debug_tool.info(f"[{self.__class__.__name__}] Loaded webpage from {file_path} successfully")
        except Exception:
            self._soup = None
            self._file_path = None
            self.debug_tool.error(f"[{self.__class__.__name__}] Failed to load webpage from {file_path}")
            raise FailedToLoadWebpageException(f"Failed to load webpage from {file_path}")

    def _read_webpage_from_file(self, file_path: (str, pathlib.Path)) -> BeautifulSoup:
        self.debug_tool.info(f"[{self.__class__.__name__}] Reading webpage from {file_path}...")
        with open(file_path, "r", encoding=self._encoding) as file:
            content = file.read()
        self.debug_tool.info(f"[{self.__class__.__name__}] Transforming {file_path} to BeautifulSoup...")
        return BeautifulSoup(content, 'lxml')

    @property
    def debug_tool(self) -> Debugger:
        return self._debug_tool

    @property
    def encoding(self) -> str:
        return self._encoding

    @property
    def file_path(self) -> pathlib.Path:
        return self._file_path

    @property
    def soup(self) -> BeautifulSoup:
        return self._soup



from wrightyrion.page_interactor.modules.handler import Handler


class ClickHandler(Handler):
    async def click_until_element_visible(self, click_selector: str, visible_selector: str, max_retry: int = 5):
        """
        Keep clicking an element until another element becomes visible or until max retries.

        :param click_selector: Selector of the element to be clicked.
        :param visible_selector: Selector of the element that should become visible after the click.
        :param max_retry: Maximum number of click attempts.
        :return: (None)
        """
        self.debug_tool.debug(f"Clicking {click_selector} until {visible_selector} becomes visible, max_retry: {max_retry}...")
        n_retry = 0
        while not await self.page.is_visible(visible_selector) and n_retry < max_retry:
            if await self.page.is_visible(click_selector):
                element_to_click = await self.page.query_selector(click_selector)
                await element_to_click.scroll_into_view_if_needed()
                await element_to_click.click()
                await self.page.wait_for_selector(visible_selector, state='visible', timeout=5000)
            else:
                self.debug_tool.error(f"Cannot find the clicking element: {click_selector}")
                raise RuntimeError(f"Cannot find the clicking element: {click_selector}")
            n_retry += 1

        if n_retry >= max_retry:
            self.debug_tool.error(f"Cannot make {visible_selector} visible after {max_retry} tries.")
            raise RuntimeError(f"Cannot make {visible_selector} visible after {max_retry} tries.")
        else:
            self.debug_tool.debug(f"Make {visible_selector} visible after {n_retry} / {max_retry} tries.")


__all__ = ['ClickHandler']



import asyncio
from typing import List, Callable

import playwright.async_api
from gembox.debug_utils import Debugger

from .common_handler import CommonHandler
from .handler import Handler


class ScrollHandler(Handler):
    def __init__(self, page: playwright.async_api.Page, debug_tool: Debugger):
        super(ScrollHandler, self).__init__(page=page, debug_tool=debug_tool)
        self._common_handler = CommonHandler(page=page, debug_tool=debug_tool)

    async def scroll_to_bottom(self, elem: playwright.async_api.ElementHandle = None):
        if elem:
            await self._page.evaluate("elem => elem.scrollTop = elem.scrollHeight", elem)
        else:
            await self._page.evaluate("window.scrollTo(0, document.body.scrollHeight)")

    async def scroll_to_top(self, elem: playwright.async_api.ElementHandle = None):
        if elem:
            await self._page.evaluate("elem => elem.scrollTop = 0", elem)
        else:
            await self._page.evaluate("window.scrollTo(0, 0)")

    async def scroll_by(self, x: int, y: int, elem: playwright.async_api.ElementHandle = None):
        if elem:
            await self._page.evaluate("args => { args.elem.scrollLeft += args.x; args.elem.scrollTop += args.y; }",
                                      {"elem": elem, "x": x, "y": y})
        else:
            await self._page.evaluate("args => window.scrollBy(args.x, args.y)", {"x": x, "y": y})

    async def scroll_to(self, x: int, y: int, elem: playwright.async_api.ElementHandle = None):
        if elem:
            await self._page.evaluate("args => { args.elem.scrollLeft = args.x; args.elem.scrollTop = args.y; }",
                                      {"elem": elem, "x": x, "y": y})
        else:
            await self._page.evaluate("args => window.scrollTo(args.x, args.y)", {"x": x, "y": y})

    async def get_scroll_height(self, elem: playwright.async_api.ElementHandle = None) -> int:
        """
        Get the scroll height of the element or the whole page.

        :param elem: ElementHandle of the specific element or None for the whole page.
        :return: Scroll height.
        """
        if elem:
            return await self._page.evaluate("elem => elem.scrollHeight", elem)
        else:
            return await self._page.evaluate("document.body.scrollHeight")

    async def get_scroll_top(self, elem: playwright.async_api.ElementHandle = None) -> int:
        """
        Get the scroll position (scrollTop) of the element or the whole page.

        :param elem: ElementHandle of the specific element or None for the whole page.
        :return: Current scroll position from the top.
        """
        if elem:
            return await self._page.evaluate("elem => elem.scrollTop", elem)
        else:
            # You can also use `window.pageYOffset` which is equivalent to `window.scrollY`.
            return await self._page.evaluate("window.scrollY")

    async def scroll_load(self,
                          scroll_step: int = 400,
                          load_wait: int = 40,
                          same_th: int = 20,
                          scroll_step_callbacks: List[Callable] = None,
                          elem: playwright.async_api.ElementHandle = None) -> None:
        """
        Scroll and load all contents, until scrolling top does not change.

        :param scroll_step: (int) The number of pixels to scroll each time. If None, scroll to bottom.
        :param load_wait: (int) The time to wait after each scroll, in milliseconds. If none, the method will wait for 100 ms
        :param same_th: (int) The threshold of the number of same scroll top to stop scrolling.
        :param scroll_step_callbacks: (List[Callable]) A callback function to be called after each scroll.
        :param elem: (playwright.async_api.ElementHandler) The element to scroll. If None, scroll the whole page.
        :return:
        """
        self.debug_tool.info(f"Scrolling and loading... scroll_step: {scroll_step}, load_wait: {load_wait}, same_th: {same_th}, elem: {elem}")
        return await self._scroll_load(scroll_step=scroll_step, load_wait=load_wait, same_th=same_th, scroll_step_callbacks=scroll_step_callbacks, elem=elem)

    async def scroll_load_selector(self,
                                   selector: str,
                                   threshold: int = None,
                                   scroll_step: int = 400,
                                   load_wait: int = 40,
                                   same_th: int = 20,
                                   scroll_step_callbacks: List[Callable] = None,
                                   log_interval: int = 100,
                                   count_check_interval: int = 5,
                                   elem: playwright.async_api.ElementHandle = None) -> List[playwright.async_api.ElementHandle]:
        """
        Scroll and load all contents, until no new content is loaded or enough specific items are collected.

        :param selector: (str) The selector of the element to scroll. If None, the method will just scroll to the bottom
        :param scroll_step: (int) The scroll step in pixels. If none, each scroll will be `scroll_to_bottom`
        :param load_wait: (int) The time to wait after each scroll, in milliseconds. If none, the method will wait for 100 ms
        :param same_th: (int) The threshold of the number of same scroll top to stop scrolling.
        :param threshold: (int) only valid when `selector` is not `None`, after loading `threshold` number of elements, the method will stop scrolling
        :param scroll_step_callbacks: (Callable) A callback function to be called after each scroll.
        :param log_interval: (int) The interval of logging the number of loaded elements.
        :param count_check_interval: (int) The interval of checking the number of elements.
        :param elem: (ElementHandle) The element to scroll. If None, scroll the whole page.

        :return: (int) The number of elements matching the selector
        """
        self.debug_tool.info(f'Scrolling and loading {selector}... threshold: {threshold}, scroll_step: {scroll_step}, load_wait: {load_wait}, same_th: {same_th}, elem: {elem}')
        await self._scroll_load(selector=selector, threshold=threshold, scroll_step=scroll_step, load_wait=load_wait,
                                same_th=same_th, scroll_step_callbacks=scroll_step_callbacks, log_interval=log_interval,
                                elem=elem, count_check_interval=count_check_interval)
        elements = await self._common_handler.get_elements(selector=selector)
        n_elements = len(elements)
        self.debug_tool.info(f'Loaded {n_elements} elements')
        return elements

    async def _scroll_load(self,
                           selector: str = None,
                           scroll_step: int = None,
                           load_wait: int = 40,
                           same_th: int = 20,
                           threshold: int = None,
                           scroll_step_callbacks: List[Callable] = None,
                           log_interval: int = 100,
                           count_check_interval: int = 5,
                           elem: playwright.async_api.ElementHandle = None):
        """
        Scroll and load all contents.

        It's very common to scroll to the bottom and wait for the page to load until no new content is loaded or enough
        specific items are collected.
        Or you just want to load the whole page.
        This method is to do that.

        :param selector: (str) the selector to listen to
        :param scroll_step: (int) the scroll step in pixels
        :param load_wait: (int) the time to wait after each scroll, in milliseconds
        :param same_th: (int) the threshold of the number of same scroll top to stop scrolling
        :param threshold: (int) only valid when `selector` is not `None`, after loading `threshold` number of elements, the method will stop scrolling
        :param scroll_step_callbacks: (List[Callable]) A callback function to be called after each scroll.
        :param log_interval: (int) The interval of logging the number of loaded elements.
        :param count_check_interval: (int) The interval of checking the number of elements.
        :param elem: (ElementHandler) The element to scroll. If None, scroll the whole page.
        :return:
        """
        same_count = 0
        last_top = None
        count_check_counter = 0
        n_selector, prev_n_selector = 0, 0
        same_sel_count, same_sel_count_th = 0, 10

        while True:
            if selector is not None:
                count_check_counter += 1

                if count_check_counter >= count_check_interval:
                    n_selector = await self._common_handler.count(selector=selector)
                    count_check_counter = 0

                    if n_selector == prev_n_selector:
                        same_sel_count += 1
                        self.debug_tool.info(f"Same selector count: {same_sel_count}, before: {prev_n_selector}, after: {n_selector}, threshold: {threshold}")
                    else:
                        self.debug_tool.info(f"Current n_selector: {n_selector}, previous n_selector: {prev_n_selector}, threshold: {threshold} , same selector count: {same_sel_count} / {same_sel_count_th}")
                        same_sel_count = 0
                        prev_n_selector = n_selector

                    if same_sel_count >= same_sel_count_th:
                        self.debug_tool.info(f"Same selector count: {same_sel_count}, same_sel_count_th: {same_sel_count_th}, stopping!! count: {n_selector}, threshold: {threshold}")
                        break

                    if threshold is not None and n_selector >= threshold:
                        self.debug_tool.info(f'Loaded {n_selector} elements, reached threshold {threshold}, stopping.')
                        break

                    elif n_selector - prev_n_selector >= log_interval:
                        self.debug_tool.info(f'Loaded {n_selector} elements so far, threshold: {threshold}.')
                        prev_n_selector = n_selector

            await self._scroll_step(elem=elem, scroll_step=scroll_step)

            # Calling callbacks
            if scroll_step_callbacks:
                for callback in scroll_step_callbacks:
                    if asyncio.iscoroutinefunction(callback):
                        await callback()
                    else:
                        callback()

            await asyncio.sleep(load_wait / 1000.)

            top = await self.get_scroll_top(elem=elem)

            if top == last_top:
                same_count += 1
                self.debug_tool.info(f"Same top count: {same_count}, same_threshold: {same_th}, top = {top}")
                if same_count >= same_th:
                    self.debug_tool.info(f'Top unchanged for {same_count} times, stopping.')
                    break
            else:
                same_count = 0

            last_top = top

    async def _scroll_step(self, scroll_step: int = None, elem: playwright.async_api.ElementHandle = None):
        """
        Scroll by `scroll_step` pixels, if scroll_step is `None`, scroll to bottom.
        """
        if scroll_step is None:
            await self.scroll_to_bottom(elem=elem)
        else:
            await self.scroll_by(0, scroll_step, elem=elem)


__all__ = ['ScrollHandler']



class BrowserNotRunningError(Exception):
    def __init__(self, message=f"Browser is not running"):
        super().__init__(message)


class NoActivePageError(Exception):
    def __init__(self, message=f"No active page found"):
        super().__init__(message)


class NonSingletonError(Exception):
    def __init__(self, message=f"Only one browser instance and one page is allowed"):
        super().__init__(message)



import pathlib
import aiofiles
from typing import Union, List, Callable, Optional

import playwright.async_api
from gembox.debug_utils import Debugger
from gembox.io import ensure_pathlib_path, check_and_make_dir

from wrightyrion.page_interactor.modules import CommonHandler, ScrollHandler, ClickHandler


class PageInteractor:
    def __init__(self, page, debug_tool: Optional[Debugger] = None):
        self._page: playwright.async_api.Page = page
        self._debug_tool = Debugger() if debug_tool is None else debug_tool
        self._common_handler = CommonHandler(page=self.page, debug_tool=self.debug_tool)
        self._scroll_handler = ScrollHandler(page=self.page, debug_tool=self.debug_tool)
        self._click_handler = ClickHandler(page=self.page, debug_tool=self.debug_tool)

    # common handler related operation
    async def get_element(self, selector: str, strict: bool = False) -> playwright.async_api.ElementHandle:
        """
        Get the element according to the selector.

        If `strict` is True, then when resolving multiple elements, the function will raise Error
        """
        return await self._common_handler.get_element(selector=selector, strict=strict)

    async def get_elements(self, selector: str) -> List[playwright.async_api.ElementHandle]:
        """
        Get the elements according to the selector.

        :param selector: (str) the selector
        :return: (List[playwright.async_api.ElementHandle]) the elements
        """
        return await self._common_handler.get_elements(selector=selector)

    async def count(self, selector: str) -> int:
        """
        Count the number of elements according to the selector.

        :param selector: (str) the selector
        :return: (int) the number of elements
        """
        return await self._common_handler.count(selector=selector)

    # Click related operation
    async def click(self, selector: str):
        return await self.page.click(selector=selector)

    async def click_until_element_visible(self, click_selector: str, visible_selector: str, max_retry: int = 5):
        """
        Keep clicking an element until another element becomes visible or until max retries.

        :param click_selector: Selector of the element to be clicked.
        :param visible_selector: Selector of the element that should become visible after the click.
        :param max_retry: Maximum number of click attempts.
        :return: (None)
        """
        return await self._click_handler.click_until_element_visible(click_selector=click_selector, visible_selector=visible_selector, max_retry=max_retry)

    # Scroll related operation
    async def scroll_to_bottom(self, elem: Optional[playwright.async_api.ElementHandle] = None):
        await self._scroll_handler.scroll_to_bottom(elem=elem)

    async def scroll_to_top(self, elem: Optional[playwright.async_api.ElementHandle] = None):
        await self._scroll_handler.scroll_to_top(elem=elem)

    async def scroll_by(self, x: int, y: int, elem: Optional[playwright.async_api.ElementHandle] = None):
        await self._scroll_handler.scroll_by(x=x, y=y, elem=elem)

    async def scroll_to(self, x: int, y: int, elem: Optional[playwright.async_api.ElementHandle] = None):
        await self._scroll_handler.scroll_to(x=x, y=y, elem=elem)

    async def scroll_load(self,
                          scroll_step: int = 400,
                          load_wait: int = 40,
                          same_th: int = 20,
                          scroll_step_callbacks: Optional[List[Callable]] = None,
                          elem: Optional[playwright.async_api.ElementHandle] = None) -> None:
        """
        Scroll and load all contents, until scrolling top does not change.

        :param scroll_step: (int) The number of pixels to scroll each time. If None, scroll to bottom.
        :param load_wait: (int) The time to wait after each scroll, in milliseconds. If none, the method will wait for 100 ms
        :param same_th: (int) The threshold of the number of same scroll top to stop scrolling.
        :param scroll_step_callbacks: (List[Callable]) A callback function to be called after each scroll.
        :param elem: (playwright.async_api.ElementHandler) The element to scroll. If None, scroll the whole page.
        :return:
        """
        return await self._scroll_handler.scroll_load(scroll_step=scroll_step, load_wait=load_wait, same_th=same_th,
                                                      scroll_step_callbacks=scroll_step_callbacks, elem=elem)

    async def scroll_load_selector(self,
                                   selector: str,
                                   threshold: Optional[int] = None,
                                   scroll_step: int = 400,
                                   load_wait: int = 40,
                                   same_th: int = 20,
                                   scroll_step_callbacks: List[Callable] = None,
                                   log_interval: int = 100,
                                   count_check_interval: int = 5,
                                   elem: Optional[playwright.async_api.ElementHandle] = None) -> List[playwright.async_api.ElementHandle]:
        """
        Scroll and load all contents, until no new content is loaded or enough specific items are collected.

        :param selector: (str) The selector of the element to scroll. If None, the method will just scroll to the bottom
        :param scroll_step: (int) The scroll step in pixels. If none, each scroll will be `scroll_to_bottom`
        :param load_wait: (int) The time to wait after each scroll, in milliseconds. If none, the method will wait for 100 ms
        :param same_th: (int) The threshold of the number of same scroll top to stop scrolling.
        :param threshold: (int) only valid when `selector` is not `None`, after loading `threshold` number of elements, the method will stop scrolling
        :param scroll_step_callbacks: (Callable) A callback function to be called after each scroll.
        :param log_interval: (int) The interval of logging the number of loaded elements.
        :param count_check_interval: (int) The interval of checking the number of loaded elements.
        :param elem: (ElementHandle) The element to scroll. If None, scroll the whole page.

        :return: (int) The number of elements matching the selector
        """
        return await self._scroll_handler.scroll_load_selector(selector=selector, threshold=threshold,
                                                               scroll_step=scroll_step, load_wait=load_wait,
                                                               same_th=same_th,
                                                               scroll_step_callbacks=scroll_step_callbacks,
                                                               log_interval=log_interval,
                                                               count_check_interval=count_check_interval, elem=elem)

    async def get_scroll_top(self, elem: playwright.async_api.ElementHandle = None) -> int:
        return await self._scroll_handler.get_scroll_top(elem=elem)

    async def get_scroll_height(self, elem: playwright.async_api.ElementHandle = None) -> int:
        return await self._scroll_handler.get_scroll_height(elem=elem)

    # Other operation
    async def type_input(self, selector: str, text: str):
        return await self.page.type(selector=selector, text=text)

    async def download_html(self,
                            file_path: Union[str, pathlib.Path],
                            elem: Optional[playwright.async_api.ElementHandle] = None,
                            encoding: str = "utf-8"):
        """
        Download the current web page or specific element.

        :param file_path: (str, pathlib.Path) the file path
        :param elem: (playwright.async_api.ElementHandle) the element to download
        :param encoding: (str, optional) the encoding of the file, default is utf-8
        :return: (None)
        """
        file_path = ensure_pathlib_path(file_path)
        check_and_make_dir(file_path.parent)

        elem_str = f"element {elem}" if elem is not None else "whole webpage"
        self.debug_tool.info(f"Downloading the {elem_str}, page url: {self.page.url}...")
        if elem is None:
            content = await self._page.content()
        else:
            content = await elem.inner_html()
        self.debug_tool.info(f"Pumping the content to {file_path}...")
        async with aiofiles.open(file_path, mode="w", encoding=encoding) as f:
            await f.write(content)
        self.debug_tool.info(f"Downloaded the {elem_str} successfully, page url: {self.page.url}, file_path: {file_path}")

    @property
    def page(self) -> playwright.async_api.Page:
        return self._page

    @property
    def debug_tool(self) -> Debugger:
        return self._debug_tool



import playwright.async_api
from gembox.debug_utils import Debugger
from playwright.async_api import async_playwright, Playwright

from .browser_mgr import SingleBrowserManager
from .page_interactor import PageInteractor
from .data_extractor import DataExtractor


class Agent:
    """
    Agent is designed to be a facade class for browser manipulation. It wraps all the interactions with the browser, and
    provides a facade for `browser_mgr`, `data_extractor` and `page_interactor`.

    - `browser_mgr` is responsible for managing the browser instance. including starting, stopping and page navigation.
    - `data_extractor` is responsible for extracting data from the page.
    - `page_interactor` is responsible for interacting with the page.

    In addition, due to the power of playwright, we also expose the playwright instance and page instance used by the agent,
    so that users can extend the functionality by themselves.

    - `self.wright` is the playwright instance.
    - `self.page` is the page instance.

    Agent has a series of hooks, which can be used to customize the agent's behavior.

    - `_init_hook` is called at the end of the `__init__` function.
    - `_start_hook` is called after the agent is started, just before setting `self._is_running` to True.
    - `_stop_hook` is called after the agent is stopped, just before setting `self._is_running` to False.
    """
    def __init__(self,
                 wright: Playwright,
                 headless: bool,
                 debug_tool: Debugger):
        """
        Initialize the Agent.

        **Note: Usually, you need to instantiate an agent by Calling `Agent.instantiate()` rather than `Agent()`**

        :param wright: (Playwright) the playwright instance
        :param headless: (bool) whether the browser is headless
        :param debug_tool: (Debugger) the debugger
        """
        assert isinstance(wright, Playwright), f"wright should be a Playwright instance, but got {wright.__class__.__name__}"
        assert isinstance(debug_tool, Debugger), f"debug_tool should be a Debugger instance, but got {debug_tool.__class__.__name__}"

        self._wright = wright
        self._headless = headless
        self._debug_tool = debug_tool
        self._browser_mgr = SingleBrowserManager(wright=wright, headless=headless, debug_tool=debug_tool)
        self._page_interactor = None
        self._data_extractor = None
        self._is_running = False

        # init hook
        self._init_hook()

    @classmethod
    async def instantiate(cls, headless=True, debug_tool=None) -> 'Agent':
        """
        Instantiate an agent.

        :param headless: (bool) whether the browser is headless
        :param debug_tool: (Debugger) the debugger
        :return: (Agent) the agent instance
        """
        wright = await (async_playwright().start())
        debug_tool = Debugger() if debug_tool is None else debug_tool
        instance = cls(wright=wright, headless=headless, debug_tool=debug_tool)
        return instance

    async def start(self, viewport: dict = None, **kwargs):
        if self.is_running:
            self.debug_tool.warn(f"{self.__class__.__name__} is already running. No need to start again")
            return

        self.debug_tool.info(f"Starting {self.__class__.__name__}...")
        await self.browser_mgr.start(viewport=viewport, **kwargs)
        self._page_interactor = PageInteractor(page=self.page, debug_tool=self.debug_tool)
        self._data_extractor = DataExtractor(page=self.page, debug_tool=self.debug_tool)
        # start hook
        await self._start_hook()
        self._is_running = True
        self.debug_tool.info(f"{self.__class__.__name__} started successfully")

    async def stop(self):
        if self.is_running is False:
            self.debug_tool.warn("Agent is not running. No need to stop")
        else:
            self.debug_tool.info(f"Stopping agent...")
            await self.browser_mgr.close()
            self._page_interactor = None
            self._data_extractor = None
            # stop hook
            await self._stop_hook()
            self._is_running = False
            self.debug_tool.info(f"Agent stopped successfully")

    # getters
    @property
    def wright(self):
        """the playwright instance"""
        return self._wright

    @property
    def browser_mgr(self) -> SingleBrowserManager:
        return self._browser_mgr

    @property
    def page_interactor(self) -> PageInteractor:
        return self._page_interactor

    @property
    def data_extractor(self) -> DataExtractor:
        return self._data_extractor

    @property
    def is_running(self) -> bool:
        return self._is_running

    @property
    def page(self) -> playwright.async_api.Page:
        return self.browser_mgr.page

    @property
    def debug_tool(self) -> Debugger:
        return self._debug_tool

    # hooks
    def _init_hook(self) -> None:
        """
        Hook function called at the end of the __init__ functions.

        When calling this hook, the agent already has `self.wright`, `self.debug_tool` and `self.browser_mgr` equipped.
        :return: (None)
        """
        pass

    async def _start_hook(self) -> None:
        """
        Hook function called after the agent is started, just before setting self._is_running to True.

        After starting the agent, the agent now have the browser instance and a blank page instance.
        :return: (None)
        """
        pass

    async def _stop_hook(self) -> None:
        """
        Hook function called after the agent is stopped, just before setting self._is_running to False.

        After stopping the agent, the agent now have no browser instance and no page instance.
        :return: (None)
        """
        pass

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.stop()


__all__ = ['Agent']



from setuptools import setup, find_packages


def read_requirements():
    with open('requirements.txt', 'r') as f:
        requirements = f.readlines()
    return [req.strip() for req in requirements if req.strip() and not req.startswith('#')]


setup(
    name='wrightyrion',
    version='0.0.20',
    packages=find_packages(exclude=['test', 'test.*']),
    install_requires=read_requirements(),
    author='steveflyer',
    author_email='steveflyer7@gmail.com',
    description='',
    # long_description=open('./README.md').read(),
    long_description_content_type='text/markdown',
)
